#!/usr/bin/env pythonTopic.releaseDb("%sPackets" %(nodeId))

import threading, thread, Queue, pickle, time, xmlrpclib, sys, cPickle, cStringIO, os, types
from Helpers import Helpers
from OFS import Folder, SimpleItem
from Globals import HTMLFile
import Globals
from ZODB import Persistence
from pyMQ import pyMQ
import urllib
from ZMessageQueueClient import Connector

def manage_addZQueueNode(context, id='ZQueueNode', nodeId='0', port=17000, REQUEST=None):
    '''Add a ZQueue Node object'''
    try:
        node = ZQueueNode(id, nodeId, port, context.absolute_url())
        context._setObject(id, node)
        if REQUEST is not None:
            return REQUEST.RESPONSE.redirect('manage_main')
        else:
            return 1
    except:
        return Helpers.GetTraceback()
    
manage_addZQueueNodeForm = HTMLFile('dtml/node_addForm',globals())

class ZQueueNode(Folder.Folder):
    '''ZQueue Node object.
    A messaging system consists of ZQueueNode objects (which may be in multiple Zopes)'''

    meta_type = "ZQueueNode"

    def log(self, message):
        try:
            s = xmlrpclib.Server('http://127.0.0.1:7050')
            s.logger.logMessage("ZQueueNodePyMQ",str(message))
        except:
            print message

    def __init__(self, id, nodeId, port, contextUrl):
        '''ZQueueNode initialiser
        id : Zope Id
        nodeId : This node's system unique id'''
        try:
            self.id = id
            self.nodeId = nodeId
            #create the pyMQ object
            dbPath = os.path.join(Globals.INSTANCE_HOME,'var')
            print dbPath
            self.port = port
            self.log(contextUrl)
##            self.xmlrpcHost = urllib.splitport(urllib.splithost(urllib.splittype(contextUrl)[1])[0])[0]
            self.xmlrpcHost = '127.0.0.1'
            self.log(self.xmlrpcHost)
            print self.port
            pyMQNode = pyMQ.MQServer()
            # self, dbPath, nodeId, hostname, port
            self.log(str((dbPath, self.nodeId, self.xmlrpcHost, self.port)))
            pyMQNode.load(dbPath, self.nodeId, self.xmlrpcHost, self.port)
            
            #set the persistently stored settings
            setPyMQNodeSettings(self.port, (dbPath, self.nodeId, self.xmlrpcHost, self.port))
            #set the cache pyMQ object
            setPyMQNode(self.port, pyMQNode)
        except:
            self.log(Helpers.GetTraceback())
            try:
                pyMQNode.stop()
            except:
                pass
            try:
                pyMQNode.unload()
            except:
                pass
            try:
                del pyMQNode
            except:
                pass
            try:
                delPyMQNodeSettings(self.port)
            except:
                pass
            try:
                delPyMQNode(self.port)
            except:
                pass
            raise
    def manage_beforeDelete(self, item=None, container=None):
        delPyMQNodeSettings(self.port)
        
#-------------------------------------------------------------------------------
#          ACCESSORS - Used by other code to get to node attributes             
#          These methods assume the correct locks have been acquired            
#-------------------------------------------------------------------------------
    def __getPyMQNode(self):
        return getPyMQNode(self.port)

    def getRemoteNodes(self):
        '''Return the list remote nodes registered with this packet
        Returns a python dict of nodeId:nodeAddress values'''
        return self.__getPyMQNode().getRemoteNodes()

    def getNodeId(self):
        '''Return this node's unique nodeId'''
        return self.nodeId

#-------------------------------------------------------------------------------
#          NODE 2 NODE Methods - Used by nodes to communicate between each other
#-------------------------------------------------------------------------------
    def needPackets(self, topicId, packetIds):
        '''Used by other nodes'''
        result = self.__getPyMQNode().needPackets(topicId, packetIds)
        print "needPackets\n%s" %result
        return result
    
    def deliverPacketsToTopic(self, nodeId, topicId, packets):
        '''Used by other nodes'''
        result = self.__getPyMQNode().deliverPacketsToTopic(nodeId, topicId, packets)
        print "deliverPacketsToTopic\n%s" %result
        return result
    
    def getRecipientsForPacket(self, topicId, packetId):
        '''Used by other nodes'''
        result = self.__getPyMQNode().getRecipientsForPacket(topicId, packetId)
        print "getRecipientsForPacket\n%s" %result
        return result
    
#-------------------------------------------------------------------------------
#          CLIENT METHODS - Used by clients that use the message queue          
#-------------------------------------------------------------------------------
    def publishMessageToTopic(self, topicId, message):
        '''Publishes a message to a topic.
        topic : case sensitive name of the topic
        message : picklable, XML-RPC marshable message body
        Creates the topic if needed.'''
        return self.__getPyMQNode().publishMessagesToTopic(topicId, None, [message])

    def subscribeToTopic(self, topicId, subscriberId,type=0):
        '''Subscribes an application to a topic.
        topic : case sensitive name of the topic
        applicationId : case sensitive name of the application
        type : for type==0, a polling client (default = 0)
        Creates the topic if needed.'''
        return self.__getPyMQNode().subscribeToTopic(topicId, subscriberId)

    def unsubscribeFromTopic(self, topicId, applicationId):
        '''Unsubscribe an application from topic.
        topic : case sensitive name of the topic
        applicationId : case sensitive name of the application
        '''
        return self.__getPyMQNode().unsubscribeFromTopic(topicId, applicationId)

    def readMessagesFromTopic(self, topicId, subscriberId, maxRead=100):
        '''Read published messages from a topic waiting for the application.
        topic : case sensitive name of the topic
        applicationId : case sensitive name of the application
        '''
        return self.__getPyMQNode().getMessagesFromTopic(topicId, subscriberId, maxRead)

    def ackMessagesFromTopic(self, topicId, subscriberId, messageIds):
        '''Acknowledge that application have received messages.
        topic : case sensitive name of the topic
        applicationId : case sensitive name of the application
        Removes the messages if no subscribers are left to read.
        This MUST be called by the application, or the messages will be re-delivered.'''
        if type(messageIds) != types.ListType:
            messageIds = [messageIds]
        return self.__getPyMQNode().acknowledgeMessagesFromTopic(topicId, subscriberId, messageIds)
#-------------------------------------------------------------------------------
#          CLIENT METHODS V2 - Updated API for client methods                   
#-------------------------------------------------------------------------------
    def publishMessagesToTopic(self, topicId, messages):
        '''Publishes message/s to a topic.
        topic : case sensitive name of the topic
        messages : picklable, XML-RPC marshable message bodies (single message or a list of)
        Creates the topic if needed.'''
        if type(messages) != type([]):
            messages = [messages]
        return self.__getPyMQNode().publishMessagesToTopic(topicId, None, messages)
    
    def getTopics(self):
        '''Get a list of available topics'''
        return self.__getPyMQNode().getTopics()
    
    def removeTopic(self, topicId):
        '''Remove a topic from the message queue'''
        self.__getPyMQNode().removeTopic(topicId)
        
#-------------------------------------------------------------------------------
# Interface related methods
#-------------------------------------------------------------------------------
    def clearNode(self, REQUEST=None):
        '''Reset the node contents.
        REQUEST : Zope request object (used for redirect)
        This will clear the node completely and re-create the ZQueuePoller object.
        Mainly used for testing purposes.'''
        pass

    def addRemoteNode(self, remoteNodeId, nodeAddress, REQUEST=None):
        '''Adds a reference to another node in the system.
        remoteNodeId : system-unique id for the remote node
        nodeAddress : XML-RPC address of the remote node (eg. http://www.foo.com/ZQueueNode0
        REQUEST : Zope request object (used for redirect)
        '''
        try:
            self.__getPyMQNode().addRemoteNode(remoteNodeId, nodeAddress)
            if REQUEST:
                self.redirect(REQUEST)
        except:
            tb = Helpers.GetTraceback()
            print tb
            raise
        
    def removeRemoteNodes(self, nodeIds, REQUEST=None):
        '''Removes references to other nodes in the system.
        nodeIds : a python list of strings containing the node Ids
        REQUEST : Zope request object (used for redirect)
        '''
        try:
            if type(nodeIds) != types.ListType:
                nodeIds = [nodeIds]
            for nodeId in nodeIds:
                self.__getPyMQNode().remRemoteNode(nodeId)
            if REQUEST:
                self.redirect(REQUEST)
        except:
            tb = Helpers.GetTraceback()
            print tb
            raise

    tableTemplate = '''<table>
    %s
    </table>'''

    #Zope Admin interface for this object, accessed through the view tab.
    def index_html(self):
        '''Renders some info'''
        try:
            nodeId = self.getNodeId()
            mqNode = self.__getPyMQNode()
            remoteNodesTable = []
            
            for key in self.getRemoteNodes().keys():
                address = self.getRemoteNodes()[key]
                remoteNodesTable.append('''<tr>
                    <td><input type="checkbox" name="nodeIds:list" value="%(key)s"></td>
                    <td>%(key)s</td><td>%(address)s</td>
                    </tr>''' %locals())            
            remoteNodesTable = self.tableTemplate %'\n'.join(remoteNodesTable)
    
            currentTopics = mqNode.getTopics()
            currentTopics = self.tableTemplate %'\n'.join(currentTopics)
                
            currentPackets = []
            # XXX remove this from rendering - not applicable any more.
            currentPackets = self.tableTemplate %'\n'.join(currentPackets)
    
            if mqNode.running:
                deliveryHandlerStatus = 'running'
            else:
                deliveryHandlerStatus = 'stopped'
    
            outputOb = HTMLFile('dtml/node',globals())
        except:
            print Helpers.GetTraceback()
            raise
        return outputOb(client=self, mapping=locals())

    def redirect(self, REQUEST):
        '''Redirect back to the referer.'''
        return REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)

import ThreadedAsync
import thread, threading

pyMQNodes = {}
pyMQNodeLock = threading.Lock()

def getPyMQNode(port):
    pyMQNodeLock.acquire()
    try:
        node = pyMQNodes[port]
        return node
    finally:
        pyMQNodeLock.release()
        
def setPyMQNode(port, node):
    pyMQNodeLock.acquire()
    try:
        node.start()
        pyMQNodes[port] = node
    finally:
        pyMQNodeLock.release()
        
def delPyMQNode(port):
    pyMQNodeLock.acquire()
    try:
        try:
            pyMQNodes[port].stop()
        except:
            pass
        del pyMQNodes[port]
    finally:
        pyMQNodeLock.release()
        
def getPyMQNodeSettings(port):        
    app = Connector.getApplication()
    return app.pyMQSettings[port]
        
def setPyMQNodeSettings(port, settings):
    app = Connector.getApplication()
    app.pyMQSettings[port] = settings
    get_transaction().commit()
    
def delPyMQNodeSettings(port):
    app = Connector.getApplication()
    del app.pyMQSettings[port]
    get_transaction().commit()
    
def getNextOpenPort():
    app = Connector.getApplication()
    keys = app.pyMQSettings.keys()
    keys.sort()
    if keys:
        return keys[-1]+1
    else:
        return 7000

def installPyMQs(map):
    def __install():
        try:
            print "INSTALLING PyMQs"
            import Zope
            app = Zope.app()
            ##Connector.getApplication()
            if not hasattr(app, 'pyMQSettings'):
                print "Creating app.pyMQSettings"
                app.pyMQSettings = Persistence.PersistentMapping()
                get_transaction().commit()
            print "app.pyMQSettings", app.pyMQSettings
            for dbPath, id, xmlrpcHost, port in app.pyMQSettings.values():
                try:
                    pyMQNode = pyMQ.MQServer()
                    pyMQNode.load(dbPath, id, xmlrpcHost, port)
                    setPyMQNode(port, pyMQNode)             
                except:
                    del app.pyMQSettings[port]
                    get_transaction().commit()
                    print "Failed creating node"
                    print Helpers.GetTraceback()
            app._p_jar.close()
            print "Done installing PyMQs"


        except:
            print "!!!ERROR INSTALLING PYMQNODES!!!"
            print Helpers.GetTraceback()
            raise
        
    #we must run __installConnectors in a thread or else the loop_callback lock will block
    thread.start_new_thread(__install, ())
ThreadedAsync.register_loop_callback(installPyMQs)