from OFS import Folder
import Message
import xmlrpclib, sys
from Helpers import Helpers
from Globals import HTMLFile
import ConflictResolvingTools
import Globals
import time, os
import ZODB
from ZODB import FileStorage

def manage_addZQueueTopic(self, id):
    '''Add a ZQueueTopic to the self folder.'''
    self._setObject(id, ZQueueTopic(id))

import threading, thread

if not hasattr(Globals,'topicDbLock'):
    Globals.topicDbLock = threading.Lock()
    Globals.topicTopicLocks = {}
    Globals.topicTopicDbs = {}

def packThread(): # XXX when happy with this, increase time between packs to an hour or so
    '''This thread packs all the db's every 2 minutes.'''
    lp = time.time()
    while 0:
        time.sleep(120)
        print "Packing dbs."
        Globals.topicDbLock.acquire()
        try:
            dbs = Globals.topicTopicDbs.keys()
        finally:
            Globals.topicDbLock.release()
        for dbName in dbs:
            db, conn, root = getDb(dbName)
            try:
                db.pack()
            finally:
                releaseDb(dbName)
        print "Packing dbs complete."

thread.start_new_thread(packThread,())

def getDb(topicId):
    '''Returns the db object.'''
    if not Globals.topicTopicDbs.has_key(topicId):
        Globals.topicDbLock.acquire()
        try:
            #connect to the db and put the db root in the dict
            dbPath = os.path.join(Globals.INSTANCE_HOME, 'Products/ZMessageQueue')
            dbFileName = os.path.join(dbPath, 'Topic%s.fs' %topicId)
            if not os.path.exists(dbPath):
                os.makedirs(dbPath)
            print "Creating db at %s" %dbPath
            db = ZODB.DB(FileStorage.FileStorage(dbFileName))
            conn = db.open()
            root = conn.root()
            Globals.topicTopicDbs[topicId] = (db,conn,root)
            Globals.topicTopicLocks[topicId] = threading.Lock()
            print "The following dbs are now available:\n%s" %'\n'.join(Globals.topicTopicLocks.keys())
        finally:
            Globals.topicDbLock.release()

    lock = Globals.topicTopicLocks[topicId]
    lock.acquire()
    return Globals.topicTopicDbs[topicId]

def releaseDb(topicId):
    '''Commit the transaction and release the db object.'''
    try:
        get_transaction().commit()
    except:
        pass
    try:
        lock = Globals.topicTopicLocks[topicId]
        lock.release()
    except KeyError:
        print "Trying to release db %s when only the following are available:\n%s" %(
            topicId, '\n'.join(Globals.topicTopicLocks.keys()))


class ZQueueTopic(Folder.Folder):
    '''ZQueue Topic Object'''
    meta_type = "ZQueueTopic"

    def __init__(self, id):
        '''Initialiser'''
        self.id = id
        self.subscribers = {}
        self.numberOfSubscribers = 0
        self.messageCounter = 0
        self.deleteCount = 0

    def getTopicUniqueId(self):
        '''Return a unique file name for this topic's database.'''
        return '%s%s' %(self.aq_parent.id, self.id)

    def getMessages(self, numberOfMessages=100):
        '''Returns all the messages contained in this topic.
        The caller MUST call releaseMessages to unlock the Topic!'''
        db,conn,root = getDb(self.getTopicUniqueId())
        ml = []
        counter = 0
        for key in root.keys():
            ml.append(root[key])
            counter += 1
            if numberOfMessages > -1 and counter > numberOfMessages:
                break
        return ml

    def releaseMessages(self):
        '''Unlocks the Topic after a getMessages call.'''
        releaseDb(self.getTopicUniqueId())

    def deleteMessages(self, messages):
        '''Deletes a list of messages.'''
        if type(messages) == type(''):
            messages = [messages]

        db,conn,root = getDb(self.getTopicUniqueId())
        try:
            for messageId in messages:
                del root[messageId]
        finally:
            releaseDb(self.getTopicUniqueId())


    def publishMessage(self, message):
        '''Publishes a message in the topic.
        message must be a valid python like object'''
        #Create a unique message Id
        messageId = "n%st%sm%s" %(self.aq_parent.getNodeId(), self.id, self.messageCounter)
        self.addMessage(messageId, message)

    def calcNumberOfSubscribers(self):
        '''Calculates the number of subscribers and stores the value.'''
        self.numberOfSubscribers = len(self.subscribers.keys())
        self._p_changed = 1

    def addMessage(self, messageId, message, fromPacket=0):
        '''Add a message to the topic.'''
        self.messageCounter += 1
        self._p_changed = 1
        if self.numberOfSubscribers or not fromPacket:
            message = Message.ZQueueMessage(messageId, message, fromPacket)
            db,conn,root = getDb(self.getTopicUniqueId())
            try:
                root[messageId] = message
                get_transaction().commit()
            finally:
                releaseDb(self.getTopicUniqueId())
##            Message.manage_addZQueueMessage(self, messageId, message, fromPacket)
        #add the message to the current publishing packet
##        if not fromPacket:
##            packet = self.aq_parent.addMessageToCurrentPacket(self.id, messageId, message)

    def subscribe(self, applicationId, type=0):
        '''Subscribes to a topic.'''
        #print "Topic: subscribe. appID:",applicationId," type: ",type
        print "1. subscribers = %s" %self.subscribers
        self.subscribers[applicationId] = type
        self.calcNumberOfSubscribers()
        self._p_changed = 1
        print "2. subscribers = %s" %self.subscribers
        #Not sure about this line - added with conflict resolving
        #packet = self.aq_parent.addMessageToCurrentPacket(self.id, messageId, message)

    def unsubscribe(self, applicationId):
        '''Unsubscribes from topic.'''
        if self.subscribers.has_key(applicationId):
            del self.subscribers[applicationId]
        self.calcNumberOfSubscribers()
        self._p_changed = 1

    def readMessages(self, applicationId, numberOfMessages=100):
        '''Read any messages waiting in topic for subscriber.
        Subscriber must ack messages to indicate that it has received them.'''
        messages = []
        counter = 0
        try:
            for message in self.getMessages():
                if not message.subscriberReceived(applicationId):
                    messages.append((message.id, message.getMessage()))
                    counter+=1
                    if numberOfMessages > -1 and counter >= numberOfMessages:
                        break
        finally:
            self.releaseMessages()
        return messages

    def ackMessages(self, applicationId, messageIds):
        '''Acknowledge successfull delivery of messages.
        This will:
            1) flag each message as being read by the subscriber.
            2) remove messages that have no more subscribers waiting for them.
        '''
        delObjects = []
        #print "topic:ack called"
        db,conn,root = getDb(self.getTopicUniqueId())
        try:
            for messageId in messageIds:
                #print "got mesageID",messageId
                message = root[messageId]
                message.addSubsriberReceived(applicationId)
                #print "self.numbverOfSubscribers:",self.numberOfSubscribers,"\n\tmessage.subreceived: ",message.getNumberOfSubscribersReceived()
                if message.getNumberOfSubscribersReceived() == self.numberOfSubscribers and message.isPackaged():
                    #print "deleting message:",messageId
                    delObjects.append(messageId)
                    #XXX a list must be kept of delivered messages (as all nodes will ask to deliver this message)

            if delObjects:
                for messageId in delObjects:
                    del root[messageId]
        finally:
            releaseDb(self.getTopicUniqueId())


    #Code to deliver messages automatically
    def deliverMessages(self):
        #print "delivering"
        for subscriber in self.subscribers.keys():
            #print "\tGot subscriber"
            if self.subscribers[subscriber] == 1:
                #print "\t\tNon-polling"
                #If not polling send msg
                messages = []
                try:
                    for message in self.getMessages():
                        #print "found msg"
                        if not message.subscriberReceived(subscriber):
                            messages.append((message.id, message.getMessage()))
                    if messages:
                        try:
                            print "Topic:Non-polling trying callback"
                            xmlrpclib.Server(subscriber).callback(self.id, messages)
                        except:
                            print "xmlrpc connction error trying to callback"
                            pass
                finally:
                    self.releaseMessages()



    def index_html(self):
        '''Admin interface for topic.'''
        topicId = self.id
        subscriberList = []
        for subscriber in self.subscribers.keys():
            subscriberList.append('''
            %s<br>''' %subscriber)

        subscriberList = "\n".join(subscriberList)

        messageList = []
        try:
            for message in self.getMessages():
                messageList.append('''
                %s<br>''' %(message.id))
        finally:
            self.releaseMessages()

        messageList = "\n".join(messageList)

        return HTMLFile('dtml/topic', globals())(client=self, mapping=locals())


    def _p_resolveConflict(self, oldState, savedState, newState):
        '''Conflict resolver.'''
        try:
            #set up a handler with a default handler that raises an error if the values are different
            handler = ConflictResolvingTools.KeyConflictHandler(ConflictResolvingTools.errorIfDifferent)
            handler.addHandleableKey('_objects', ConflictResolvingTools.tupleMerge)
            handler.addHandleableKey('messageCounter', ConflictResolvingTools.returnLargest)
            handler.addHandleableKey('subscribers', ConflictResolvingTools.dictMerge)
            handler.addHandleableKey('numberOfSubscribers', ConflictResolvingTools.returnLargest)
            retState = ConflictResolvingTools.mergeDiffs(oldState, savedState, newState, handler)
            retState['numberOfSubscribers'] = len(retState['subscribers'].keys())
            #make sure that there are keys for every id in _objects.  Remove those that are not relevant
            ConflictResolvingTools.synchroniseObjects(retState)
            return retState
        except:
            tb = Helpers.GetTraceback()
            print "!" * 120
            print "Error resolving conflict for Topic"
            print tb
            return None
