import sys
import os
import time
import xmlrpclib
import threading
import types
import cStringIO
import cPickle
import threading
from Products.ZMessageQueue.Helpers import Helpers
import ThreadedAsync
import urllib

import AccessControl.SecurityManagement, AccessControl.User
import ZODB, ZODB.ZApplication
import Globals, OFS.Application

class ZopeApplicationWrapper:
    '''Zope application wrapper that releases the connection once the
    object is dereferenced.'''
    def __init__(self):
        try:
            import Globals
            DB = Globals.DB
            # "Log on" as system user
            AccessControl.SecurityManagement.newSecurityManager(
                None, AccessControl.User.system)
            # Set up the "application" object that automagically opens
            # connections
            app=ZODB.ZApplication.ZApplicationWrapper(
                DB, 'Application', OFS.Application.Application, (),
                Globals.VersionNameName)
            # "Log off" as system user
            AccessControl.SecurityManagement.noSecurityManager()
            self.connection = DB.open(temporary=1)
            self.logToTimelineLogger("Getting self.app")
            self.app = app(self.connection)
            self.logToTimelineLogger("Got self.app")
        except:
            self.logToTimelineLogger("Could not create self.app:\n%s" %Helpers.GetTraceback())

    def __getattr__(self, attrName):
        return getattr(self.app, attrName)
    
    def __del__(self):
        pass
##        self.logToTimelineLogger("Releasing self.app")
##        try:
##            self.connection.close()
##            #self.app._p_jar.close()
##        except:
##            self.logToTimelineLogger("Error releasing self.app:\n%s" %Helpers.GetTraceback())
##        del self.app

    def logToTimelineLogger(self, message):
        try:
            s = xmlrpclib.Server('http://127.0.0.1:7050')
            s.logger.logMessage("%sZopeApplicationWrapper" %(
                os.getpid()), "%s : %s" %(threading.currentThread().getName(),message))
        except:
            print message

        
def getApplication():
    '''Get a Zope Application wrapper object'''
    return ZopeApplicationWrapper()

def getPollApplication():
    return ZopeApplicationWrapper()

class Connector(threading.Thread):
    '''Connector client for ZMessageQueue'''
    def __init__(self, nodeAddress, applicationId, pollInterval=1):
        '''Initialiser.'''
        threading.Thread.__init__(self)
        self.setDaemon(1)
        self.__applicationId = applicationId
        self.__nodeAddress = nodeAddress
        self.__callbackMethodsDict = {}
        self.__stopEvent = threading.Event()
        self.__pollInterval = pollInterval
        self.__logger = xmlrpclib.Server('http://127.0.0.1:7050')
        #register with the Zope main loop to fire us up once Zope is runnning
        ThreadedAsync.register_loop_callback(self.asyncCallback)
        self.log("Connector __init__ done")
        
    def asyncCallback(self, map):
        '''Fire our own loop up.'''
        self.log("FIRING UP CONNECTOR")
        self.start()        
        
    def stop(self):
        '''Stop the polling cycle.'''
        self.log("STOPPING CONNECTOR")
        self.__stopEvent.set()
        
    def log(self, message):
        try:
            s = xmlrpclib.Server('http://127.0.0.1:7050')
            s.logger.logMessage(self.__applicationId, message)
        except:
            print message
        
    def run(self):
        '''Message queue polling loop.
        checks for messages on all topics and calls callback methods if 
        such methods were registered when subscribe was called.'''
        try:
            while not self.__stopEvent.isSet():
                print self.__applicationId,
                for topicId in self.__callbackMethodsDict.keys():
                    #get the messages
                    messages = self.readMessagesFromTopic(topicId)
                    if messages:
                        try:
                            app = getPollApplication()
                            #do a callback for messages
                            ackList = []
                            appUrl, methodName = self.__callbackMethodsDict[topicId]
                            appUrl = urllib.splithost(urllib.splittype(appUrl)[1])[1]
                            try:
                                ob = app.restrictedTraverse(appUrl)
                                method = getattr(ob, methodName)
                                ackList = method(messages)
                                self.log("Dispatched messages:\n%s" %("\n".join([str(message.__dict__) for id,message in messages])))
                            except:
                                print "CONNECTOR CALLBACK FAILED"
                                print "CALLBACK DETAILS:\nURL = %s\nMethod = %s" %(appUrl, methodName)
                                print Helpers.GetTraceback()
                            if ackList:
                                self.ackMessagesFromTopic(topicId, ackList)
                        finally:
                            try:
                                #this is so that we release the connection.
                                del app
                            except:
                                pass
    
                self.__stopEvent.wait(self.__pollInterval)
        except:
            self.log("ERROR:%s" %Helpers.GetTraceback())
            raise

        
    def __handleXMLRPCFault(self, e):
        print "."
        exception = e.faultString.split()[-1].strip()
        if exception not in ['ServerBusyError']:#,'NotFound:']:
            print "="*60+"\nError Handling XML-RPC Fault\n" + "="*60 + "\n"
            print "FAULTSTRING:"
            print "------------"
            print e.faultString
            print "Traceback:"
            print "----------"
            print Helpers.GetTraceback()
            print "="*60+"\n"
            raise e
    
    def getServer(self):
        '''Return the xml-rpc server proxy'''
        try:
            return xmlrpclib.Server(self.__nodeAddress)
        finally:
            pass
##            self.log("Got Connector")
            
    
    marshallerTypes = xmlrpclib.Marshaller.dispatch.keys()
    def publishMessageToTopic(self, topic, message):
        '''Publishes a message to a topic'''
        #if the message is an instance, pickle the instance and send a Binary
        if type(message) not in self.marshallerTypes or type(message) == types.InstanceType:
            f = cStringIO.StringIO()
            cPickle.dump(message, f)
            f.seek(0)
            message = xmlrpclib.Binary(f.read())
        done = 0
        while not done:
            try:
                self.getServer().publishMessageToTopic(topic, message)
                done = 1
            except xmlrpclib.Fault, e:
                self.__handleXMLRPCFault(e)
                time.sleep(0.25)
        self.log("PUBLISHING to %s done" %topic)

    def subscribeToTopic(self, topic, callbackObjectUrl=None, 
                        callbackMethodName=None):
        '''Subscribe to a topic.
        topic : id of the topic to subscribe to
        type : Not used
        callbackMethod : method that must be called back if messages arrives in a topic
            callbackMethod must accept a list of (messageId, message) tuples and return
            a list of messageIds that were processed successfully.
            callbackMethod may not raise an exception - it will crash the auto polling loop.'''
        self.log("A")
        if type(callbackObjectUrl) in [types.IntType, types.FloatType]:
            typ = callbackObjectUrl
            callbackObjectUrl = None
        self.log("B")
            
        typ = 0
        done = 0
        while not done:
            self.log("C")
            try:
                self.log("C.1")
                s = self.getServer()
                self.log("C.2 : (%s,%s,%s)" %(topic, self.__applicationId,typ))
                s.subscribeToTopic(topic, self.__applicationId,typ)
                self.log("C.3")
                done = 1
            except xmlrpclib.Fault, e:
                self.__handleXMLRPCFault(e)
                time.sleep(0.25)
        self.log("D")
        if callbackObjectUrl is not None and callbackMethodName is not None:
            self.log("E")
            self.__callbackMethodsDict[topic] = (
                callbackObjectUrl, callbackMethodName)
        self.log("SUBSCRIPTION to %s done" %topic)

    def unsubscribeFromTopic(self, topic):
        '''Unsubscribes from a topic - also removes auto callbacks.'''
        done = 0
        while not done:
            try:
                self.getServer().unsubscribeFromTopic(topic, self.__applicationId)
                done = 1
            except xmlrpclib.Fault, e:
                self.__handleXMLRPCFault(e)
                time.sleep(0.25)
        if self.__callbackMethodsDict.has_key(topic):
            del self.__callbackMethodsDict[topic]
        self.log("UNSUBSCRIPTION from %s done" %topic)
            
    def readMessagesFromTopic(self, topic, numberOfMessages=100):
        '''Read messages waiting in topic
        Returns a list of (messageId, message) tuples'''
        done = 0
        while not done:
            try:
                rawMessages = self.getServer().readMessagesFromTopic(topic, self.__applicationId, numberOfMessages)
                messages = [(id, msg) for id,msg in rawMessages 
                            if not isinstance(msg, xmlrpclib.Binary)]
                messages += [(id, cPickle.load(cStringIO.StringIO(msg.data))) for id,msg in rawMessages 
                            if isinstance(msg, xmlrpclib.Binary)]
                done = 1
            except xmlrpclib.Fault, e:
                self.__handleXMLRPCFault(e)
                time.sleep(0.25)
        return messages

    def ackMessagesFromTopic(self, topic, messageIds):
        '''Acknowledge receipt of messages - must be a list of messageIds'''
        done = 0
        while not done:
            try:
                self.getServer().ackMessagesFromTopic(topic, self.__applicationId, messageIds)
                done = 1
            except xmlrpclib.Fault, e:
                self.__handleXMLRPCFault(e)
                time.sleep(0.25)
