from time import time
from time import sleep
from constants import *
import threading
import thread
import types
import copy

class NoWaitException:pass
class LockTimeOutException:pass
class MaxKeepTimeOutException:pass

def MaxTimeout():
    raise MaxKeepTimeOutException

class ReadOnlyShell:
    def __init__(self, owner, changeMethodList = []):
        self.__changeMethodList = changeMethodList
    

class acquireReadLock:
    def __init__(self, parent, lock, ignoreacquireChangeLock = 0, changeMethodList = []):
        self.__changeMethodList = changeMethodList
        self.__lock = lock
        self.__ignoreacquireChangeLock = ignoreacquireChangeLock
        self.__parent = parent
        if not self.__ignoreacquireChangeLock:
            self.__parent._Lock__statThread.acquire()
            try:
                self.__parent._Lock__currentRead+=1
            finally:
                self.__parent._Lock__statThread.release()
    def __del__(self):
        if not self.__ignoreacquireChangeLock:
            self.__parent._Lock__statThread.acquire()
            try:
                self.__parent._Lock__currentRead-=1
            finally:
                self.__parent._Lock__statThread.release()
        self.__lock.release()
    def owner(self):
        return self.__parent._Lock__owner

class expiryReadLock(acquireReadLock):
    def __init__(self, parent, lock, timeout, ignoreacquireChangeLock = 0, changeMethodList = []):
        acquireReadLock.__init__(self, parent, lock, ignoreacquireChangeLock, changeMethodList)
        self.__timeoutThread = timeout
        self.__timeoutThread.start()
    def __del__(self):
        self.__timeoutThread.cancel()
        acquireReadLock.__del__(self)

class acquireChangeLock:
    def __init__(self, parent, changeLock, locks):
        self.__locks = locks
        self.__changeLock = changeLock
        self.__parent = parent
        self.__parent._Lock__statThread.acquire()
        try:
            self.__parent._Lock__currentChange+=1
        finally:
            self.__parent._Lock__statThread.release()
    def __del__(self):
        self.__parent._Lock__statThread.acquire()
        try:
            self.__parent._Lock__currentChange-=1
        finally:
            self.__parent._Lock__statThread.release()
        if self.__changeLock.locked():
            self.__changeLock.release()
    def release(self):
        if hasattr(self, '__locks'):
            del self.__locks
        if self.__changeLock.locked():
            self.__changeLock.release()
    def owner(self):
        return self.__parent._Lock__owner

class expiryChangeLock(acquireChangeLock):
    def __init__(self, parent, changeLock, locks, timeout):
        acquireChangeLock.__init__(self, parent, changeLock, locks)
        self.__timeoutThread = timeout
        self.__timeoutThread.start()
    def __del__(self):
        self.__timeoutThread.cancel()
        acquireChangeLock.__del__(self)

class Lock:
    def __init__(self, owner, maxReadSlots, timeout = NEVER, changeMethodList = []):
        self.__maxReadSlots = maxReadSlots
        self.__timeout = timeout
        self.__changeMethodList = changeMethodList
        self.__currentRead = 0
        self.__currentChange = 0

        self.__lockArray = threading.Semaphore(maxReadSlots)

        self.__acquireReadLockThread = thread.allocate_lock()
        self.__acquireChangeLockThread = thread.allocate_lock()
        self.__changeBusyLockThread = thread.allocate_lock()
        self.__statThread = thread.allocate_lock()

        self.__simultaneousWaits = 0
        self.__maxSimultaneousWaits = 0
        self.__maxWait = 0
        self.__maxWaitChange = 0

        self.__owner = owner

    def maxReadSlots(self):
        return self.__maxReadSlots
    
    def increaseCurrentRead(self):
        self.__currentRead+=1
    
    def increaseCurrentChange(self):
        self.__currentChange+=1

    def decreaseCurrentRead(self):
        self.__currentRead-=1
    
    def decreaseCurrentChange(self):
        self.__currentChange-=1

    def currentRead(self):
        return self.__currentRead

    def currentChange(self):
        return self.__currentChange

    def simultaneousWaits(self):
        self.__statThread.acquire()
        try:
            return self.__simultaneousWaits
        finally:
            self.__statThread.release()

    def maxWait(self):
        return self.__maxWait

    def maxWaitChange(self):
        return self.__maxWaitChange

    def acquireReadLock(self, timeout = None, ignoreacquireChangeLock = 0):
        timeStamp = time()
        self.__statThread.acquire()
        try:
            self.__simultaneousWaits+=1
        finally:
            self.__statThread.release()
        try:
            lastTimeStamp = timeStamp
            if timeout != None:
                while self.__acquireChangeLockThread.locked() and not ignoreacquireChangeLock:
                    timeout = timeout - (time() - lastTimeStamp)
                    lastTimeStamp = time()
                    if timeout <= 0:
                        raise LockTimeOutException
                    sleep(0.001)
            else:
                while self.__acquireChangeLockThread.locked() and not ignoreacquireChangeLock:
                    sleep(0.1)
            if timeout == None:
                self.__acquireReadLockThread.acquire()
            else:
                while 1:
                    if self.__acquireReadLockThread.acquire(1):
                        break
                    else:
                        timeout = timeout - (time() - lastTimeStamp)
                        lastTimeStamp = time()
                        if timeout <= 0:
                            raise LockTimeOutException
        finally:
            self.__statThread.acquire()
            try:
                if self.__maxSimultaneousWaits < self.__simultaneousWaits:
                    self.__maxSimultaneousWaits = self.__simultaneousWaits
                self.__simultaneousWaits-=1
            finally:
                self.__statThread.release()
            
        try:
            while 1:
                lock = self.__lockArray.acquire(0)
                if lock:
                    break
                elif self.__changeBusyLockThread.locked() and ignoreacquireChangeLock:
                    sleep(0.001)
                    return None
                elif timeout != None:
                    timeout = timeout - (time() - lastTimeStamp)
                    lastTimeStamp = time()
                    if timeout <= 0:
                        raise LockTimeOutException
                sleep(0.001)
            if self.__timeout == NEVER:
                return acquireReadLock(self, self.__lockArray, ignoreacquireChangeLock, self.__changeMethodList)
            else:
                timer = threading.Timer(self.__timeout, MaxTimeout)
                return expiryReadLock(self, self.__lockArray, timer, ignoreacquireChangeLock, self.__changeMethodList)
        finally:
            maxWait = time() - timeStamp
            if maxWait > self.__maxWait:
                self.__maxWait = maxWait
            self.__acquireReadLockThread.release()

    def acquireChangeLock(self, timeout = None):
        try:
            # Initialize
            locks = []
            timeStamp = time()
            lastTimeStamp = timeStamp
            acquiredBusyLock = 0
            acquiredLock = 0
    
            # If no timeout
            if timeout == None:
                acquiredBusyLock = self.__changeBusyLockThread.acquire()
                acquiredLock = self.__acquireChangeLockThread.acquire()
                try:
                    while len(locks) != self.maxReadSlots():
                        while 1:
                            lock = self.acquireReadLock(ignoreacquireChangeLock = 1)
                            if lock:
                                locks+=[lock]
                                break
                            sleep(0.001)
                finally:
                    self.__changeBusyLockThread.release()
                    acquiredBusyLock = 0
    
            # If timeout
            else:
                while 1:
                    acquiredBusyLock = self.__changeBusyLockThread.acquire(1)
                    if acquiredBusyLock:
                        break
                    else:
                        timeout = timeout - (time() - lastTimeStamp)
                        lastTimeStamp = time()
                        if timeout <= 0:
                            raise LockTimeOutException
                while 1:
                    acquiredLock = self.__acquireChangeLockThread.acquire(1)
                    if acquiredLock:
                        break
                    elif timeout <= 0:
                        raise LockTimeOutException
                    timeout = timeout - (time() - lastTimeStamp)
                    lastTimeStamp = time()
                while len(locks) != self.maxReadSlots():
                    while 1:
                        timeout = timeout - (time()-lastTimeStamp)
                        lastTimeStamp = time()
                        if timeout <= 0:
                            raise LockTimeOutException
                        else:
                            lock = self.acquireReadLock(timeout = timeout, ignoreacquireChangeLock = 1)
                            if lock:
                                locks+=[lock]
                                break
                            sleep(0.001)
                self.__changeBusyLockThread.release()
                acquiredBusyLock = 0
            if len(locks) == self.maxReadSlots():
                maxWait = time() - timeStamp
            if maxWait > self.__maxWaitChange:
                self.__maxWaitChange = maxWait
            if self.__timeout == NEVER:
                return acquireChangeLock(self, self.__acquireChangeLockThread, self.__lockArray)
            else:
                timer = threading.Timer(self.__timeout, MaxTimeout)
                return expiryChangeLock(self, self.__acquireChangeLockThread, self.__lockArray, timer)
        except:
            if acquiredBusyLock:
                self.__changeBusyLockThread.release()
            if acquiredLock:
                self.__acquireChangeLockThread.release()
            raise




