from ZODB import Persistence
from time import time
from Lock import *
import thread
import threading
import traceback
import copy
import xmlrpclib
import socket
from ZODB import PersistentList

class SimpleObject(Persistence.Persistent):
    def addNewAttr(self, object, name):
        self.__dict__[name] = object
        self._p_changed = 1

class Topic(Persistence.Persistent):
    def __init__(self, owner, topicId, persistent = 0):
        self.__topicId = topicId
        self.__persistent = persistent
        self.__timeStamp = time()
        
        self.__subscribers = Persistence.PersistentMapping()
        self.__messages = Persistence.PersistentMapping()
        self.__unreadMessages = Persistence.PersistentMapping()
        self.__unpacketedMessages = PersistentList.PersistentList()
        self.__messageIdCounter = 0
        
        self.__highPriorityQueue = SimpleObject()
        self.__highPriorityQueue.idCounter = 0
        self.__normalPriorityQueue = SimpleObject()
        self.__normalPriorityQueue.idCounter = 0
        self.__lowPriorityQueue = SimpleObject()
        self.__lowPriorityQueue.idCounter = 0
        
        self.__highPriorityQueueWaiting = PersistentList.PersistentList()
        self.__normalPriorityQueueWaiting = PersistentList.PersistentList()
        self.__lowPriorityQueueWaiting = PersistentList.PersistentList()
        
        self.__packets = Persistence.PersistentMapping()
        self.__packetHistory = Persistence.PersistentMapping()
        
        self.__packetIdCounter = 0
        
        self.__noNodeDeliveryWaitPeriod = 30
        
        self.start(owner)
    
    def getNodeId(self):
        return self._v_owner.getNodeId()
    
    def setNoNodeDeliveryWaitPeriod(self, newValue):
        self.__noNodeDeliveryWaitPeriod = newValue
        return 1
    
    def getNoNodeDeliveryWaitPeriod(self):
        return self.__noNodeDeliveryWaitPeriod
    
    def getTopicId(self):
        return self.__topicId
    
    def getRemoteNodes(self):
        return self._v_owner.getRemoteNodes()
    
    ###########################################################
    #                                                         #
    ###########################################################
    
    def start(self, owner):
        try:
            self._v_owner = owner
            self.__started = 1
            self._v_subscribers = Lock(None, 5)
            self._v_messages = Lock(None, 5)
            self._v_unreadMessages = Lock(None, 5)
            self._v_unpacketedMessages = Lock(None, 5)
            self._v_messageIdCounter = Lock(None, 1)
            
            self._v_packets = Lock(None, 2)
            self._v_packetHistory = Lock(None, 1)
            self._v_packetIdCounter = Lock(None, 1)
            
            self._v_highPriorityQueue = Lock(None, 1)
            self._v_normalPriorityQueue = Lock(None, 1)
            self._v_lowPriorityQueue = Lock(None, 1)
        
            self._v_highPriorityQueueWaiting = Lock(None, 2)
            self._v_normalPriorityQueueWaiting = Lock(None, 2)
            self._v_lowPriorityQueueWaiting = Lock(None, 2)
        
            self._v_highPriorityQueueEvent = threading.Event()
            self._v_normalPriorityQueueEvent = threading.Event()
            self._v_lowPriorityQueueEvent = threading.Event()
            self._v_processPacketsEvent = threading.Event()
    
            self._v_packetManagementEvent = threading.Event()
    
            self._v_normalPriorityQueueEvent.set()
            self._v_packetManagementEvent.set()
            
            self._v_sleepNode = {}
            
            thread.start_new(self._bpNormalPriorityMessages, ())
            thread.start_new(self._bpPacketManager, ())
            thread.start_new(self._controller, ())
            thread.start_new(self._processPackets, ())

            self._v_nodeTimeOut = threading.Event()
        except:
            print 'Exception in start'
            traceback.print_exc()
            raise
    
    def stop(self):
        try:
            self.__started = 0
            self._v_lowPriorityQueueEvent.clear()
            self._v_normalPriorityQueueEvent.clear()
            self._v_highPriorityQueueEvent.clear()
            self._v_packetManagementEvent.clear()
            self._v_processPacketsEvent.clear()
            del self._v_lowPriorityQueueEvent
            del self._v_normalPriorityQueueEvent
            del self._v_highPriorityQueueEvent
            del self._v_packetManagementEvent
            del self._v_processPacketsEvent
            del self._v_subscribers
            del self._v_messages
            del self._v_unreadMessages
            del self._v_unpacketedMessages
            del self._v_messageIdCounter
            del self._v_packets
            del self._v_packetHistory
            del self._v_packetIdCounter
            del self._v_highPriorityQueue
            del self._v_normalPriorityQueue
            del self._v_lowPriorityQueue
        except:
            print 'Exception in stop'
            traceback.print_exc()
            raise

    def processNormalQueue(self, queueId):
        try:
            unreadMessagesLock = self._v_unreadMessages.acquireChangeLock()
            unpacketedMessagesLock = self._v_unpacketedMessages.acquireChangeLock()
            messagesLock = self._v_messages.acquireChangeLock()
            lock = self._v_normalPriorityQueueWaiting.acquireChangeLock()
            
            try:
                queue = getattr(self.__normalPriorityQueue, queueId)
            except:
                self.__normalPriorityQueueWaiting.remove(queueId)
                return
            header = PersistentList.PersistentList([len(queue.subscribers), queue.packeted, time(), queue.publisherId])
            for messageId in queue.messages.keys():
                self.__messages[messageId] = PersistentList.PersistentList([queue.messages[messageId]]) + header
            for subscriberId in queue.subscribers:
                self.__unreadMessages[subscriberId] += queue.messages.keys()
            if not queue.packeted:
                self.__unpacketedMessages += queue.messages.keys()
            self.__normalPriorityQueueWaiting.remove(queueId)
            delattr(self.__normalPriorityQueue, queueId)
        except:
            print 'Exception in processNormalQueue'
            traceback.print_exc()
            raise
        
    def _bpNormalPriorityMessages(self):
        try:
            emptyTries = 1
            try:
                while self.__started:
                    self._v_normalPriorityQueueEvent.wait()
                    sleep(0.1)
                    lock = self._v_normalPriorityQueueWaiting.acquireReadLock()
                    waiting = copy.copy(list(self.__normalPriorityQueueWaiting))
                    del lock
                    if self.__started and waiting:
                        self._v_lowPriorityQueueEvent.clear()
                        for queueId in waiting:
                            self.processNormalQueue(queueId)
                        get_transaction().commit()
                    elif not self.__started:
                        break
                    else:
                        emptyTries -= 1
                        if not emptyTries:
                            emptyTries = 3
                            self._v_normalPriorityQueueEvent.clear()
                            #self._v_packetManagementEvent.set()
                            self._v_lowPriorityQueueEvent.set()
                    sleep(0.20)
            except:
                import traceback
                print '_bpNormalPriorityMessages is BROKEN!!!'
                traceback.print_exc()
        except:
            print 'Exception in _bpNormalPriorityMessages'
            traceback.print_exc()
            raise
    
    def _bpPacketManager(self):
        try:
            emptyTries = 2
            try:
                while self.__started:
                    self._v_packetManagementEvent.wait()
                    lockunpacketedMessages = self._v_unpacketedMessages.acquireChangeLock()
                    lockpacketIdCounter = self._v_packetIdCounter.acquireChangeLock()
                    lockpackets = self._v_packets.acquireChangeLock()
                    total = len(self.__unpacketedMessages)
                    messageIds = self.__unpacketedMessages[:100]
                    del lockunpacketedMessages
                    if self.__started and messageIds:
                        nodeId = self.getNodeId()
                        self.__unpacketedMessages = self.__unpacketedMessages[100:]
                        packetId = nodeId+'.'+self.__topicId+'.'+`self.__packetIdCounter`
                        self.__packetIdCounter += 1
                        # packetId originaterId nodeId messageIdsOnly messages/messageIds
                        if len(self.getRemoteNodes().keys()):
                            packet = PersistentList.PersistentList([packetId, nodeId, [nodeId], 1, []])
                            for messageId in messageIds:
                                #self.__messages[messageId][2] = 1 Move this to deliver function
                                packet[4] += [messageId]
                            self.__packets[packetId] = packet
                            del packet
                        get_transaction().commit()
                        #self._v_processPacketsEvent.set() XXXXX
                    elif not self.__started:
                        break
                    else:
                        emptyTries -= 1
                        if not emptyTries:
                            emptyTries = 2
                            self._v_packetManagementEvent.clear()
                            self._v_processPacketsEvent.set()
                    del lockpackets
                    del lockpacketIdCounter
                    sleep(0.25)
            except Exception, e:
                import traceback
                print e
                for line in traceback.extract_stack():
                    print line
        except:
            print 'Exception in _bpPacketManager'
            traceback.print_exc()
            raise
    
    ###########################################################
    #                                                         #
    ###########################################################
    
    def _controller(self):
        try:
            packCheck = 5
            while self.__started:
                if not packCheck:
                    packCheck = 4
                    self._v_packetManagementEvent.set()
                else:
                    packCheck -= 1
                if not self._v_normalPriorityQueueEvent.isSet():
                    self._v_normalPriorityQueueEvent.set()
                sleep(2)
        except:
            print 'Exception in _controller'
            traceback.print_exc()
            raise
    
    ###########################################################
    #                                                         #
    ###########################################################
    
    def subscribed(self, subscriberId):
        try:
            subscriberLock = self._v_subscribers.acquireReadLock()
            subscribers = self.__subscribers
            return subscribers.has_key(subscriberId)
        except:
            print 'Exception in subscribed'
            traceback.print_exc()
            raise

    def subscribe(self, subscriberId, callback = ''):
        try:
            subscriberLock = self._v_subscribers.acquireChangeLock()
            unreadMessagesLock = self._v_messages.acquireChangeLock()
            if subscriberId in self.__subscribers.keys():
                return 0
            else:
                self.__subscribers[subscriberId] = callback
                self.__unreadMessages[subscriberId] = PersistentList.PersistentList()
                return 1
        except:
            print 'Exception in subscribe'
            traceback.print_exc()
            raise

    def unsubscribe(self, subscriberId):
        try:
            subscriberLock = self._v_subscribers.acquireChangeLock()
            unreadMessagesLock = self._v_unreadMessages.acquireChangeLock()
            if subscriberId not in self.__subscribers.keys():
                return 0
            else:
                del self.__subscribers[subscriberId]
                del self.__unreadMessages[subscriberId]
                return 1
        except:
            print 'Exception in unsubscribe'
            traceback.print_exc()
            raise
    
    def getMessageIds(self, count):
        try:
            lock = self._v_messageIdCounter.acquireChangeLock()
            result = [i for i in range(self.__messageIdCounter, self.__messageIdCounter+count)]
            self.__messageIdCounter = result[-1] + 1
            return result
        except:
            print 'Exception in getMessageIds'
            traceback.print_exc()
            raise
    
    def newNormalPriorityQueue(self):
        try:
            lock = self._v_normalPriorityQueue.acquireChangeLock()
            id = self.__normalPriorityQueue.idCounter
            self.__normalPriorityQueue.idCounter += 1
            result = SimpleObject()
            self.__normalPriorityQueue.addNewAttr(result, 'temp_%s' % id)
            return (result, 'temp_%s' % id)
        except:
            print 'Exception in newNormalPriorityQueue'
            traceback.print_exc()
            raise
    
    def publishMessages(self, publisherId, messages, packeted = 0, priority = 1):
        try:
            if priority == 1:
                messageIds = self.getMessageIds(len(messages))
                queue, name = self.newNormalPriorityQueue()
                lock = self._v_subscribers.acquireReadLock(5)
                queue.subscribers = self.__subscribers.keys()
                del lock
                queue.messages = Persistence.PersistentMapping()
                for i in range(len(messages)):
                    queue.messages[messageIds[i]] = messages[i]
                queue.packeted = packeted
                queue.publisherId = publisherId
                lock = self._v_normalPriorityQueueWaiting.acquireChangeLock()
                self.__normalPriorityQueueWaiting += PersistentList.PersistentList([name])
                del lock
                return messageIds
            else:
                raise "Invalid priority"
        except:
            print 'Exception in publishMessages'
            traceback.print_exc()
            raise
        
    def getMessages(self, subscriberId, maxRead = 0, otherHeaders = 0):
        try:
            lock = self._v_subscribers.acquireReadLock()
            if not self.__subscribers.has_key(subscriberId):
                del lock
                self.subscribe(subscriberId)
            else:
                del lock
            messagesLock = self._v_messages.acquireReadLock()
            unreadMessagesLock = self._v_unreadMessages.acquireReadLock()
            if not otherHeaders:
                result = []
                unread = self.__unreadMessages[subscriberId]
                if maxRead:
                    unread = unread[:maxRead]
                for id in unread:
                    result += [(id, self.__messages[id][0])]
                return result
            else:
                raise "NotImplimented"
        except:
            print 'Exception in getMessages'
            traceback.print_exc()
            raise
        
    def acknowledgeMessages(self, subscriberId, messageIds):
        try:
            count = 0
            messagesLock = self._v_messages.acquireChangeLock()
            unreadMessagesLock = self._v_unreadMessages.acquireChangeLock()
            subscriberLock = self._v_subscribers.acquireReadLock()
            if not self.__subscribers.has_key(subscriberId):
                del subscriberLock
                self.subscribe(subscriberId)
            else:
                del subscriberLock
                for messageId in messageIds:
                    try:
                        self.__unreadMessages[subscriberId].remove(messageId)
                        self.__messages[messageId][1] -= 1
                        count += 1
                    except:
                        pass
            return count
        except:
            print 'Exception in acknowledgeMessages'
            traceback.print_exc()
            raise
    
    def getAllPacketsForProcessing(self, timeout = 30):
        try:
            if self.__started:
                lock = self._v_packets.acquireChangeLock(timeout)
                result = self.__packets
                self.__packets = Persistence.PersistentMapping()
                del lock
                return result
            else:
                return []
        except:
            print 'Exception in getAllPacketsForProcessing'
            traceback.print_exc()
            raise
    
    def getPacketIds(self, timeout = 30):
        try:
            lock = self._v_packets.acquireReadLock(timeout)
            result = self.__packets.keys()
            #self.__packets = Persistence.PersistentMapping()
            del lock
            return result
        except:
            print 'Exception in getPacketIds'
            traceback.print_exc()
            raise
    
    def getPacketsUsingIds(self, packetIds, timeout = 30):
        try:
            finishedPublishing = 0
            while not finishedPublishing:
                finishedPublishing = 1
                for packetId in packetIds:
                    lock = self._v_packets.acquireReadLock()
                    packet = self.__packets[packetId]
                    if not self.hasMessages(packet[4]):
                        finishedPublishing = 0
                        #print 'not finishedPublishing'
                        del lock
                        sleep(0.1)
                        break
                    del lock
            lock = self._v_packets.acquireReadLock(timeout)
            ownPacketIds = self.__packets.keys()
            result = []
            for packetId in packetIds:
                if packetId in ownPacketIds:
                    packet = copy.copy(list(self.__packets[packetId]))
                    if packet[3]:
                        messageLock = self._v_messages.acquireReadLock()
                        messages = []
                        for messageId in packet[4]:
                            messages += [self.__messages[messageId][0]]
                        del messageLock
                        packet[4] = messages
                        packet[3] = 0
                        if not self.__packets[packetId][3]:
                            print 'OOPSY!'
                    result += [packet]
            del lock
            return result
        except:
            print 'Exception in getPacketsUsingIds'
            print '\n'*5
            print 'packet[4]:'
            print packet[4]
            print
            print 'self.__messages.keys()', self.__messages.keys()
            print 
            traceback.print_exc()
            raise
    
    def deliverPackets(self, nodeId, packets, timeout = 30):
        try:
            if self.__started:
                lock = self._v_packets.acquireChangeLock(timeout)
                result = []
                for packet in packets:
                    if packet[0] not in self.__packets.keys():
                        print 'I am adding packet', packet[0], 'from node', nodeId, 'to my list.'
                        messageIds = self.publishMessages('Node:'+nodeId, packet[4], 1)
                        #allMessageIdsAccountedFor = 0
                        #while not allMessageIdsAccountedFor:
                        #    sleep(0.1)
                        #    allMessageIdsAccountedFor = 1
                        #    messageLock = self._v_messages.acquireReadLock()
                        #    keySnapShot = self.__messages.keys()
                        #    del messageLock
                        #    for messageId in messageIds:
                        #        if messageId not in keySnapShot:
                        #            allMessageIdsAccountedFor = 0
                        #            print 'Not All accounted for'
                        #            break
                        packet[4] = messageIds
                        packet[3] = 1
                        nodeId = self.getNodeId()
                        if nodeId not in packet[2]:
                            packet[2] += [nodeId]
                        self.__packets[packet[0]] = packet
                        result += [packet[0]]
                    else:
                        print 'I already have packet', packet[0]
                get_transaction().commit()
                return result
            else:
                return []
        except:
            print 'Exception in deliverPackets'
            traceback.print_exc()
            raise
    
    def updateRecipientsForpacket(self, packetId, recipients, timeout = 30):
        try:
            if self.__started:
                lock = self._v_packets.acquireChangeLock(timeout)
                result = 0
                if self.__packets.has_key(packetId):
                    for recipient in recipients:
                        if recipient not in self.__packets[packetId][2]:
                            self.__packets[packetId][2] += [recipient]
                            result += 1
                del lock
                return result
            else:
                return 0
        except:
            print 'Exception in updateRecipientsForpacket'
            traceback.print_exc()
            raise
    
    def needPackets(self, packetIds):
        try:
            lockpackets = self._v_packets.acquireReadLock()
            lockpacketHistory = self._v_packetHistory.acquireReadLock()
            ownPacketIds = self.__packets.keys()
            ownPacketIds += self.__packetHistory.keys()
            del lockpackets
            del lockpacketHistory
            result = [packetId for packetId in packetIds if packetId not in ownPacketIds]
            #for packetId in packetIds:
            #    if packetId not in ownPacketIds:
            #        result += [packetId]
            return result
        except:
            print 'Exception in needPackets'
            traceback.print_exc()
            raise
    
    def getRecipientsForPacket(self, packetId):
        try:
            packet = None
            lock = self._v_packets.acquireReadLock()
            lock2 = self._v_packetHistory.acquireReadLock()
            if self.__packets.has_key(packetId):
                packet = self.__packets[packetId]
                return packet[2]
            if not packet:
                if self.__packetHistory.has_key(packetId):
                    packet = self.__packetHistory[packetId]
                    return packet
            return []
        except:
            print 'Exception in getRecipientsForPacket'
            traceback.print_exc()
            raise

    def archivePacket(self, packetId):
        try:
            lockpackets = self._v_packets.acquireChangeLock()
            lockpacketHistory = self._v_packetHistory.acquireChangeLock()
            if self.__packets.has_key(packetId):
                self.__packetHistory[packetId] = self.__packets[packetId][2]
                del self.__packets[packetId]
                return 1
            else:
                return 0
        except:
            print 'Exception in archivePacket'
            traceback.print_exc()
            raise
    
    def hasMessages(self, messageIds):
        lock = self._v_messages.acquireReadLock()
        snapshot = self.__messages.keys()
        del lock
        for messageId in messageIds:
            if messageId not in snapshot:
                return 0
        return 1
    
    def _processNodePackets(self, nodeAddress, topicId, packetIds, remoteNodes, nodeId):
        try:
            connection = xmlrpclib.Server(nodeAddress)
            connected = 0
            try:
                wanted = connection.needPackets(topicId, packetIds)
                connected = 1
            except:
                pass
            if connected:
                if wanted:
                    print 'Node', nodeAddress, 'wanted the following packets:'
                    for packetId_ in wanted:
                        print '\t', packetId_
                    packets = self.getPacketsUsingIds(wanted)
                    result = connection.deliverPacketsToTopic(nodeId, topicId, packets)
                wanted = connection.needPackets(topicId, packetIds)
                for packetId in packetIds:
                    if packetId not in wanted:
                        recipients = connection.getRecipientsForPacket(topicId, packetId)
                        result = self.updateRecipientsForpacket(packetId, recipients)
                        packetRecipients = self.getRecipientsForPacket(packetId)
                        pendingNodes = 0
                        for node in remoteNodes.keys():
                            if node not in packetRecipients:
                                pendingNodes += 1
                        if not pendingNodes:
                            print 'No more nodes need packet', packetId
                            self.archivePacket(packetId)
                        else:
                            print pendingNodes, ' nodes need packet', packetId
                print 'finished _processNodePackets'
            else:
                print 'Could not make initial connection with remote node', nodeAddress
        except:
            print 'Exception in _processNodePackets'
            traceback.print_exc()
            raise
    
    def _processPackets(self):
        try:
            nodeId = self.getNodeId()
            topicId = self.getTopicId()
            emptyTries = 2
            print 'Starting background process _processPackets'
            try:
                while self.__started:
                    self._v_processPacketsEvent.wait(0.25)
                    if not self.__started:
                        return
                    lockPackets = self._v_packets.acquireReadLock()
                    packetIds = self.__packets.keys()
                    del lockPackets
                    if packetIds:
                        print 'Not all remote nodes recieved these packets:'
                        for packetId_ in packetIds:
                            print '\t', packetId_
                        remoteNodes = self.getRemoteNodes()
                        print 'remoteNodes:', remoteNodes
                        count = len(remoteNodes.keys())
                        maxCount = count
                        for nodeId in remoteNodes.keys():
                            nodeAddress = remoteNodes[nodeId]
                            if not self._v_sleepNode.has_key(nodeId):
                                try:
                                    self._processNodePackets(nodeAddress, topicId, packetIds, remoteNodes, nodeId)
                                    get_transaction().commit()
                                    count -= 1
                                except:
                                    try:
                                        get_transaction().abort()
                                        print 'ERROR in _processPackets'
                                        traceback.print_exc()
                                        print '======= Will retry node later ======='
                                        self._v_sleepNode[nodeId] = 100
                                    except:
                                        print 'Serious error in _processPackets'
                            else:
                                self._v_sleepNode[nodeId] -= 1
                                if not self._v_sleepNode[nodeId]:
                                    del self._v_sleepNode[nodeId]
                                sleep(0.01)
                        if maxCount == count:
                            print 'Could not deliver to any remote Nodes,'
                            print ' so, I\'ll be sleeping for %d seconds.' % self.__noNodeDeliveryWaitPeriod
                            sleep(self.__noNodeDeliveryWaitPeriod)
                    else:
                        emptyTries -= 1
                        if not emptyTries:
                            emptyTries = 2
                            self._v_processPacketsEvent.clear()
                            #print '_v_processPacketsEvent is going to sleep'
                    sleep(0.01)
            except Exception, e:
                print '_bpNormalPriorityMessages is BROKEN!!!'
                traceback.print_exc()
                print '======================================'
        except:
            print 'Exception in _processPackets'
            traceback.print_exc()
            raise
    
