##############################################################################
#
# Copyright (c) 2003 GMS and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the GNU GENERAL PUBLIC LICENSE
# Version 2 of June 1991 (GPL).
#
##############################################################################
"""
This is version 0.0.0.1 of the pyMQ software package.
"""

from QueueXMLRPCServer import *
from constants import *
import thread
from Lock import *
from time import sleep
from time import time
from time import asctime
from ZODB import FileStorage
from ZODB import MappingStorage
from ZODB import DB
from ZODB import Transaction
from ZODB import Persistence
from Topic import Topic
import xmlrpclib
import traceback
import os
from ZODB.POSException import ConflictError, ReadConflictError


class NodeWasNotLoadedException: pass

class pyMQHandler(QueueHandler):
    def exp_addTopic(self, topicId, persistent = 0):
        result = self.server.parent().addTopic(topicId, persistent)
        return result
    def exp_removeTopic(self, topicId):
        result = self.server.parent().removeTopic(topicId)
        return result
    def exp_getTopics(self):
        result = self.server.parent().getTopics()
        return result
    def exp_subscribeToTopic(self, topicId, subscriberId, callbackInfo = ''):
        result = self.server.parent().subscribeToTopic(topicId, subscriberId, callbackInfo)
        return result
    def exp_unsubscribeFromTopic(self, topicId, subscriberId):
        result = self.server.parent().unsubscribeFromTopic(topicId, subscriberId)
        return result
    def exp_subscribedToTopic(self, topicId, subscriberId):
        result = self.server.parent().subscribedToTopic(topicId, subscriberId)
        return result
    def exp_publishMessagesToTopic(self, topicId, publisherId, messages):
        result = self.server.parent().publishMessagesToTopic(topicId, publisherId, messages)
        return result
    def exp_getMessagesFromTopic(self, topicId, subscriberId, maxRead = 0, otherHeaders = 0):
        result = self.server.parent().getMessagesFromTopic(topicId, subscriberId, maxRead, otherHeaders)
        return result
    def exp_acknowledgeMessagesFromTopic(self, topicId, subscriberId, messageIds):
        result = self.server.parent().acknowledgeMessagesFromTopic(topicId, subscriberId, messageIds)
        return result
    def exp_needPackets(self, topicId, packetIds):
        return self.server.parent().needPackets(topicId, packetIds)
    def exp_deliverPacketsToTopic(self, nodeId, topicId, packets):
        result = self.server.parent().deliverPacketsToTopic(nodeId, topicId, packets)
        return result
    def exp_getRecipientsForPacket(self, topicId, packetId):
        result = self.server.parent().getRecipientsForPacket(topicId, packetId)
        return result
    def exp_hasTopicMessages(self, topicId, messageIds):
        result = self.server.parent().hasTopicMessages(topicId, messageIds)
        return result
    def exp_restart(self, startDelay = 0):
        result = self.server.parent().restart(startDelay)
        return result
    def exp_flushQueue(self):
        try:
            return 1
        finally:
            thread.start_new(self.server.parent().flushQueue(), ())
            del self

class MQServer:
    def restart(self, startDelay = 0):
        try:
            return 1
        finally:
            self.stop()
            sleep(startDelay)
            self.start()
    def load(self, dbPath, nodeId, hostname, port):
        print 'Starting up Node', nodeId, 'on port', port
        self.__dbPath = dbPath
        self.__storage = FileStorage.FileStorage(dbPath + '\\' + nodeId)
        self.__db = DB(self.__storage, 56, 3200)
        self.__conn = self.__db.open()
        self.__root = self.__conn.root()
        self.__nodeId = nodeId
        self.__port = port
        self.__hostname = hostname
        
        self.running = 0
        
        if not self.__root.has_key('topics'):
            self.__root['topics'] = Persistence.PersistentMapping()
            self.__root['config'] = Persistence.PersistentMapping()
            self.__root['packets'] = Persistence.PersistentMapping()
            self.__root['packetHistory'] = Persistence.PersistentMapping()
            self.__root['packetsIdCounter'] = 0
            self.__root['remoteNodes'] = Persistence.PersistentMapping()
            print 'rooting', self.__root
        
        self.__remoteNodes = self.__root['remoteNodes']
        
        get_transaction().commit()
        
        self.setup()
        
        self.loaded = 1
    
    def unload(self):
        if not hasattr(self, 'loaded'):
            raise NodeWasNotLoadedException
        print 'Unloading Node', self.getNodeId()

        if self.running:
            self.stop()
        
        self.__conn.close()
        self.__db.close()
        self.__storage.close()
        
        del self.loaded
        del self.__remoteNodes
        del self.__root
        
        del self.__dbPath
        del self.__hostname
        del self.__port
        del self.__nodeId
        
        del self.__conn
        del self.__db
        del self.__storage
        
        return 1
    
    def flushQueue(self):
        started = 0
        if not self.loaded:
            raise QueueMustBeLoadedError
        dbPath = self.__dbPath
        nodeId = self.__nodeId
        hostname = self.__hostname
        port = self.__port
        print 'Flushing queue...'
        if self.running:
            print 'stopping queue...'
            self.stop()
            started = 1
        print 'unloading queue...'
        self.unload()
        fileName = dbPath + '\\' + nodeId
        if os.path.exists(fileName):
            print 'removing "%s"...' % fileName
            os.remove(fileName)
        if os.path.exists(fileName + '.tmp'):
            print 'removing "%s"...' % (fileName + '.tmp')
            os.remove(fileName + '.tmp')
        if os.path.exists(fileName + '.lock'):
            print 'removing "%s"...' % (fileName + '.lock')
            os.remove(fileName + '.lock')
        if os.path.exists(fileName + '.index'):
            print 'removing "%s"...' % (fileName + '.index')
            os.remove(fileName + '.index')
        if os.path.exists(fileName + '.old'):
            print 'removing "%s"...' % (fileName + '.old')
            os.remove(fileName + '.old')
        print 'loading queue...'
        self.load(dbPath, nodeId, hostname, port)
        if started:
            print 'starting queue...'
            self.start()
        print 'Done!\n'
    
    def getNodeId(self):
        return self.__nodeId

    def hasRemoteNode(self, nodeId):
        return self.__remoteNodes.has_key(nodeId)
    
    def addRemoteNode(self, nodeId, remoteAddress):
        self.__remoteNodes[nodeId] = remoteAddress
        get_transaction().commit()
        return 1
    
    def getRemoteNodes(self):
        return self.__remoteNodes
    
    def remRemoteNode(self, nodeId):
        del self.__remoteNodes[nodeId]
        get_transaction().commit()
    
    def setup(self):
        '''The Queue Node setup is done here.'''
        pass
    
    def _packDB(self):
        while self.running:
            try:
                self.__db.pack()
                if os.path.exists(self.__dbPath+'.old'):
                    os.remove(self.__dbPath+'.old')
            except FileStorage.FileStorageError:
                pass
            except Exception, e:
                print 'pyMQ._packDB: %s' % str(e)
            for x in range(120):
                if not self.running:
                    return
                sleep(0.5)
    
    def _processPackets(self):
        status = 0
        nodeId = self.getNodeId()
        while self.running:
            for topic in self.__root['topics'].values():
                topicId = topic.getTopicId()
                try:
                    lock = topic.getPackets()
                    packetIds = topic.getPacketIds()
                    del lock
                    if status:
                        print 'Processing', packetIds
                    if packetIds:
                        for nodeAddress in self.__remoteNodes.values():
                            try:
                                if status:
                                    print 'Connecting to', nodeAddress
                                connection = xmlrpclib.Server(nodeAddress)
                                want = connection.needPackets(topicId, packetIds)
                                if want:
                                    packets = topic.getPacketsUsingIds(want)
                                    result = connection.deliverPacketsToTopic(nodeId, topicId, packets)
                                want = connection.needPackets(topicId, packetIds)
                                for packetId in packetIds:
                                    if packetId not in want:
                                        recipients = connection.getRecipientsForPacket(topicId, packetId)
                                        result = topic.updateRecipientsForpacket(packetId, recipients)
                                    packetRecipients = topic.getRecipientsForPacket(packetId)
                                    remoteNodes = self.getRemoteNodes()
                                    pendingNodes = 0
                                    for node in remoteNodes:
                                        if node not in packetRecipients:
                                            pendingNodes += 1
                                    if not pendingNodes:
                                        topic.archivePacket(packetId)
                                    get_transaction().commit()
                            except Exception, e:
                                status = 1
                                get_transaction().abort()
                                print 'ERROR in _processPackets'
                                traceback.print_exc()
                                print '======= Will retry node later ======='
                except Exception, e:
                    print 'ERROR in _processPackets'
                    traceback.print_exc()
                    print '======= Will retry topic later ======='
                sleep(0.5)
            sleep(0.5)
    
    def start(self):
        try:
            if self.running:
                raise "Queue is already running"
            else:
                self.topicLock = Lock(None, 20)
                self.packets = self.__root['packets']
                self.packetHistory = self.__root['packetHistory']
                self.packetHistoryLock = Lock(None, 3)
                self.packetsLock = Lock(None, 3)
                self.packetsIdCounterLock = Lock(None, 3)
                for x in range(100):
                    try:
                        self.XMLRPC = QueueXMLRPCServer(self, (self.__hostname, self.__port), pyMQHandler)
                        break
                    except:
                        import os
                        #os.system('cls')
                        print 'Could not create QueueXMLRPCServer, retrying...'
                        sleep(0.25)
                        if x == 99:
                            raise
                for topic in self.__root['topics'].values():
                    topic.start(self)
                thread.start_new(self.XMLRPC .run, ())
                self.running = 1
                thread.start_new(self._packDB, ())
            return 1
        except:
            traceback.print_exc()
            raise
    
    def stop(self):
        try:
            if not self.running:
                raise "Queue was not started."
            else:
                self.XMLRPC.stop()
                self.XMLRPC.server_close()
                self.XMLRPC.socket.close()
                while not self.XMLRPC.stopped:
                    sleep(0.1)
                    print '...',
                print
                print '\n'*5
                print dir(self.XMLRPC)
                print '\n'*5
                print
                del self.XMLRPC
                del self.packets
                for topic in self.__root['topics'].values():
                    topic.stop()
                topicLock = self.topicLock.acquireChangeLock()
                packetsLock = self.packetsLock.acquireChangeLock()
                packetsIdCounterLock = self.packetsIdCounterLock.acquireChangeLock()
                del self.topicLock
                del self.packetsLock
                del self.packetsIdCounterLock
                del topicLock
                del packetsLock
                del packetsIdCounterLock
                self.running = 0
            return 1
        except:
            traceback.print_exc()
            raise
    
    def addTopic(self, topicId, lock = 1, persistent = 0):
        conflict = 1
        while conflict:
            try:
                conflict = 0
                if lock:
                    lock = self.topicLock.acquireChangeLock()
                root = self.__root
                topics = root['topics']
                if topics.has_key(topicId):
                    raise "This topic already exists."
                else:
                    topics[topicId] = Topic(self, topicId, persistent)
                get_transaction().commit()
                return 1
            except ConflictError:
                get_transaction().abort()
                print 'Experienced a conflict error:', ConflictError
                conflict = 1

    def removeTopic(self, topicId):
        conflict = 1
        while conflict:
            try:
                conflict = 0
                lock = self.topicLock.acquireChangeLock()
                root = self.__root
                topics = root['topics']
                if not topics.has_key(topicId):
                    raise "This topic does not exist."
                else:
                    del topics[topicId]
                print self.__root
                get_transaction().commit()
                return 1
            except ConflictError:
                get_transaction().abort()
                print 'Experienced a conflict error:', ConflictError
                conflict = 1

    def getTopics(self):
        root = self.__root
        try:
            lock = self.topicLock.acquireReadLock()
            topics = root['topics']
            return topics.keys()
        except:
            import traceback
            traceback.print_exc()
            raise
    
    def subscribeToTopic(self, topicId, subscriberId, callbackInfo = ''):
        conflict = 1
        while conflict:
            try:
                conflict = 0
                root = self.__root
                lock = self.topicLock.acquireReadLock()
                topics = root['topics']
                if not topics.has_key(topicId):
                    self.addTopic(topicId, 0)
                topic = topics[topicId]
                result = topic.subscribe(subscriberId, callbackInfo)
                get_transaction().commit()
                return result
            except ConflictError:
                get_transaction().abort()
                print 'Experienced a conflict error:', ConflictError
                conflict = 1

    def unsubscribeFromTopic(self, topicId, subscriberId):
        conflict = 1
        while conflict:
            try:
                conflict = 0
                root = self.__root
                lock = self.topicLock.acquireReadLock()
                topics = root['topics']
                if not topics.has_key(topicId):
                    self.addTopic(topicId, 0)
                topic = topics[topicId]
                result = topic.unsubscribe(subscriberId)
                get_transaction().commit()
                return result
            except ConflictError:
                get_transaction().abort()
                print 'Experienced a conflict error:', ConflictError
                conflict = 1

    def subscribedToTopic(self, topicId, subscriberId):
        try:
            root = self.__root
            lock = self.topicLock.acquireReadLock()
            topics = root['topics']
            if not topics.has_key(topicId):
                return 0
            topic = topics[topicId]
            result = topic.subscribed(subscriberId)
            return result
        except:
            traceback.print_exc()
            raise
    
    def publishMessagesToTopic(self, topicId, publisherId, messages, packeted = 0, priority = 1):
        conflict = 1
        while conflict:
            try:
                conflict = 0
                try:
                    root = self.__root
                    lock = self.topicLock.acquireReadLock()
                    topics = root['topics']
                    if not topics.has_key(topicId):
                        self.addTopic(topicId, 0)
                    topic = topics[topicId]
                    result = topic.publishMessages(publisherId, messages)
                    get_transaction().commit()
                    return result
                except:
                    traceback.print_exc()
                    raise
            except ConflictError:
                get_transaction().abort()
                print 'Experienced a conflict error:', ConflictError
                conflict = 1

    def getMessagesFromTopic(self, topicId, subscriberId, maxRead = 0, otherHeaders = 0):
        root = self.__root
        lock = self.topicLock.acquireReadLock()
        topics = root['topics']
        if not topics.has_key(topicId):
            self.addTopic(topicId, 0)
        topic = topics[topicId]
        result = topic.getMessages(subscriberId, maxRead, otherHeaders)
        return result
    
    def acknowledgeMessagesFromTopic(self, topicId, subscriberId, messageIds):
        conflict = 1
        while conflict:
            try:
                conflict = 0
                root = self.__root
                lock = self.topicLock.acquireReadLock()
                topics = root['topics']
                if not topics.has_key(topicId):
                    self.addTopic(topicId, 0)
                topic = topics[topicId]
                result = topic.acknowledgeMessages(subscriberId, messageIds)
                get_transaction().commit()
                return result
            except ConflictError:
                get_transaction().abort()
                print 'Experienced a conflict error:', ConflictError
                conflict = 1

    def needPackets(self, topicId, packetIds):
        root = self.__root
        lock = self.topicLock.acquireReadLock()
        topics = root['topics']
        if not topics.has_key(topicId):
            self.addTopic(topicId, 0)
        topic = topics[topicId]
        result = topic.needPackets(packetIds)
        return result
    
    def addRemotePackets(self, nodeId, packets, priority = 1):
        lock = self._v_packets.acquireReadLock()
        ownPacketIds = self.__packets.keys()
        del lock
        for packet in packets:
            if packet not in ownPacketIds:
                messages = packet[4]
                messageIds = self.publishMessages('Node ' + nodeId, messages, 1, priority)
    
    def deliverPacketsToTopic(self, nodeId, topicId, packets):
        conflict = 1
        while conflict:
            try:
                conflict = 0
                root = self.__root
                lock = self.topicLock.acquireReadLock()
                topics = root['topics']
                if not topics.has_key(topicId):
                    self.addTopic(topicId, 0)
                topic = topics[topicId]
                result = topic.deliverPackets(nodeId, packets)
                get_transaction().commit()
                return result
            except ConflictError:
                get_transaction().abort()
                print 'Experienced a conflict error:', ConflictError
                conflict = 1

    def getRecipientsForPacket(self, topicId, packetId):
        try:
            root = self.__root
            lock = self.topicLock.acquireReadLock()
            topics = root['topics']
            if not topics.has_key(topicId):
                self.addTopic(topicId, 0)
            topic = topics[topicId]
            result = topic.getRecipientsForPacket(packetId)
            return result
        except:
            traceback.print_exc()
            raise
    
    def hasTopicMessages(self, topicId, messageIds):
        root = self.__root
        lock = self.topicLock.acquireReadLock()
        topics = root['topics']
        if not topics.has_key(topicId):
            self.addTopic(topicId, 0)
        topic = topics[topicId]
        result = topic.hasMessages(messageIds)
        return result


class pyMQ:
    pass
